from __future__ import annotations
from tmt.storage.schema import Entry, Result
from tmt.storage.json_db import DbManager
from uuid import uuid4
from datetime import datetime
from contextvars import ContextVar
from tmt.configs.parser import Configs
from typing import Optional, Any
from tmt.exceptions import DuplicatedNameError
import os
import sys
import pickle


context_manager: ContextVar[ContextManager] = ContextVar('context_manager', default=None)


class ContextManager:
    def __init__(self, name: str, config_path: Optional[str] = None, allow_duplicate_names=False):
        if config_path:
            self.config = Configs.from_config(config_path)
        else:
            self.config = Configs.default_config()
        if DbManager(self.config.json_db_path).get_entries_by_name(name) and not allow_duplicate_names:
            raise DuplicatedNameError(f'one (or more) entry with name {name} already exists. Set `allow_duplicate_names=True` or change name (recommended)')
        self.entry = Entry(
            id=str(uuid4()), 
            name=name, 
            args=' '.join(sys.argv), 
            date_created=int(datetime.now().timestamp()), 
            local_results_path=self.config.results_path,
            )
        self.snap_manager = self.config.init_snapshot_manager(self.entry.id)
        self.entry.local_snapshot_path = self.snap_manager.snapshot_dest
        os.makedirs(self.get_save_path(), exist_ok=True)
        self.last_saved_counter = 0

    def save(self, obj: Any, name: str, allow_exist=False, extension='.pkl') -> str:
        path = self.get_save_path_with_name(name)
        if os.path.exists(path):
            if not allow_exist:
                raise DuplicatedNameError(f'{self.get_save_path_with_name(name)} already exists and `allow_exist` is False. Specify another name to save this object')
            else:
                path = self.increment_last_saved_path(name)
        path += extension
        with open(path, 'wb') as f:
            pickle.dump(obj, f)
        self.entry.results.append(Result(self.entry.id, name, path))
        return path

    def get_save_path_with_name(self, name: str):
        return os.path.join(self.get_save_path(), name)

    def get_save_path(self):
        return os.path.join(self.entry.local_results_path, self.entry.id)

    def increment_last_saved_path(self, name: str):
        self.last_saved_counter += 1
        return os.path.join(self.get_save_path(), f'{name}_{self.last_saved_counter}')



from prompt_toolkit.layout.containers import Window, HSplit, VSplit, to_container, HorizontalAlign, WindowAlign
from prompt_toolkit.layout.controls import BufferControl
from prompt_toolkit.widgets import Frame, Button, Label, Shadow
from prompt_toolkit.layout import ScrollablePane
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.utils import Event
from typing import Optional, Callable, Iterable
from tmt.configs.parser import Configs
from tmt.storage.json_db import DbManager
from tmt.storage.schema import Entry
from tmt.interface.tui.base import BaseApp, FocusableText, date_formatter
from functools import partial
from enum import Enum, auto


class SearchBy(Enum):
    NAME = auto()
    ID = auto()
    DATE = auto()


class SearchLayout(BaseApp):

    def __init__(self, back_handler: Callable, entry_select_handler: Callable, config_path: Optional[str] = None):
        if config_path:
            self.config = Configs.from_config(config_path)
        else:
            self.config = Configs.from_default_path_or_default_config()
        self.db = DbManager(self.config.json_db_path, read_only=True)
        self.results_box = HSplit([Label(text='No results...')])
        self.search_label = Label(text='', dont_extend_width=True)
        self.search_text = Buffer(multiline=False)
        self.search_box = VSplit([self.search_label, Window(BufferControl(self.search_text))], padding=2)
        self.back_handler = back_handler
        self.entry_select_handler = entry_select_handler

    def search_by_name_layout(self, clear):
        self.search_label.text = 'Search name:'
        self.search_text.on_text_changed = Event(self.search_text, self.search_by_name)
        if clear:
            self.__clear_results()
            self.__clear_search_text()
        return self.__search_layout()

    def search_by_id_layout(self, clear):
        self.search_label.text = 'Search ID (exact match):'
        self.search_text.on_text_changed = Event(self.search_text, self.search_by_id)
        if clear:
            self.__clear_results()
            self.__clear_search_text()
        return self.__search_layout()

    def search_by_name(self, buffer: Buffer):
        try:
            entries = self.db.get_entries_by_name_regex(buffer.text)
        except:
            entries = []

        if buffer.text == '' or len(entries) == 0:
            self.__clear_results()
            return
        self._show_results(entries)

    def search_by_id(self, buffer: Buffer):
        entry = self.db.get_entry_by_id(buffer.text)
        if buffer.text == '' or entry is None:
            self.__clear_results()
            return
        self._show_results((entry,))
    
    def search_by_date(self, buffer: Buffer):
        pass
        
    def _show_results(self, entries: Iterable[Entry]):
        self.results_box.children = []
        ids = HSplit([Label(text='ID')])
        names = HSplit([Label(text='Name')])
        dates = HSplit([Label(text='Date created')])
        self.results_box.children.append(VSplit([ids, names, dates]))
        for entry in entries:
            ids.children.append(to_container(FocusableText(entry.id, handler=partial(self.entry_select_handler, entry))))
            names.children.append(to_container(Label(text=entry.name)))
            dates.children.append(to_container(Label(text=date_formatter(entry.date_created))))


    def _bottom_toolbar(self):
        toolbar = super()._bottom_toolbar()
        toolbar.children.insert(0, to_container(Button('back', handler=self.back_handler)))
        return toolbar

    def __clear_results(self):
        self.results_box.children = [to_container(Label(text='No results...'))]

    def __clear_search_text(self):
        self.search_text.reset()

    def __search_layout(self):
        return HSplit([Frame(body=HSplit([self.search_box, Shadow(Frame(ScrollablePane(self.results_box)))], padding=1)), self._bottom_toolbar()], padding=5)




from prompt_toolkit.layout.containers import Window, HSplit, VSplit, to_container, Float, FloatContainer
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout import ScrollablePane
from prompt_toolkit.widgets import Frame, Button, Box, Label, Shadow
from prompt_toolkit.clipboard.pyperclip import PyperclipClipboard
from prompt_toolkit.formatted_text import HTML
from tmt.interface.tui.base import BaseApp, FocusableText, date_formatter
from tmt.storage.schema import Entry
from typing import Callable
from functools import partial
from prompt_toolkit.application.current import get_app


class EntryDisplay(BaseApp):
    def __init__(self, entry: Entry, back_handler: Callable):
        self.entry = entry
        self.back_handler = back_handler
        self.entry_disp = Frame(title=f'Experiment: {entry.name}', body=self.entry_layout())
        self.root_container = HSplit([self.entry_disp, self._bottom_toolbar()], padding=3)

    def get_layout(self):
        return self.root_container

    def entry_layout(self):
        keys = HSplit([
            Label(text='ID:', dont_extend_width=True),
            Label(text='Name:', dont_extend_width=True),
            Label(text='Args:', dont_extend_width=True),        
            Label(text='Date created:', dont_extend_width=True),
            Label(text='Metrics:', dont_extend_width=True),
            Label(text='Results path:', dont_extend_width=True),
            Label(text='Code snapshot path <enter to copy>:', dont_extend_width=True),
            Label(text='Date saved:', dont_extend_width=True)
        ], padding=1)
        ft_metric = FocusableText(text='<enter> to expand', handler=None)
        metric_float = FloatContainer(content=ft_metric, floats=[])
        ft_metric.set_handler(partial(self._on_metric_expand, metric_float))

        ft_results = FocusableText(text='<enter> to expand', handler=None)
        results_float = FloatContainer(content=ft_results, floats=[])
        ft_results.set_handler(partial(self._on_results_expand, results_float))
        vals = HSplit([
            Label(text=self.entry.id, dont_extend_width=True),
            Label(text=self.entry.name, dont_extend_width=True),
            Label(text=f'{self.entry.args or "NA"}', dont_extend_width=True),
            Label(text=date_formatter(float(self.entry.date_created)), dont_extend_width=True),
            metric_float,
            results_float,
            FocusableText(text=f'{self.entry.local_snapshot_path or "NA"}', handler=lambda: PyperclipClipboard().set_text(self.entry.local_snapshot_path)),
            Label(text=date_formatter(float(self.entry.date_saved)) if self.entry.date_saved else 'NA', dont_extend_width=True)
        ], padding=1)
        return Box(VSplit([keys, vals], padding=5))

    def _on_metric_expand(self, metric_float):
        metric_float.floats = [Float(Shadow(Frame(ScrollablePane(self._metric_layout(metric_float)))), top=0, width=20, height=10)]
        get_app().layout.focus(metric_float)

    def _on_metric_close(self, metric_float):
        metric_float.floats = []
        get_app().layout.focus(metric_float)

    def _on_results_expand(self, results_float):
        results_float.floats = [Float(Shadow(Frame(ScrollablePane(self._results_layout(results_float)))), left=0, width=40, height=10)]
        get_app().layout.focus(results_float)

    def _on_results_close(self, results_float):
        results_float.floats = []
        get_app().layout.focus(results_float)

    def _metric_layout(self, metric_float):
        keys = []
        vals = []
        if len(self.entry.metrics) == 0:
            keys.append(Label(text='No metrics found'))
        for metric in self.entry.metrics:
            keys.append(FocusableText(text=f'{metric.name.capitalize()}:'))
            vals.append(Window(content=FormattedTextControl(text=f'{metric.value:.5f}'), wrap_lines=True))
        return self._expandable_layout(keys, vals, metric_float, self._on_metric_close)

    def _results_layout(self, results_float):
        keys = []
        vals = []
        if len(self.entry.results) == 0:
            keys.append(Label(text='No results found'))
        keys.append(Label(text=HTML('<b>&lt;enter&gt; to copy</b>')))
        vals.append(Label(text=''))
        for res in self.entry.results:
            keys.append(FocusableText(text=f'{res.name.capitalize()} path:', handler=lambda: PyperclipClipboard().set_text(res.path)))
            vals.append(Window(content=FormattedTextControl(text=res.path), wrap_lines=True))
        return self._expandable_layout(keys, vals, results_float, self._on_results_close)

    def _expandable_layout(self, keys, vals, floating, handler):
        return HSplit([VSplit([HSplit(keys, padding=1), HSplit(vals, padding=1)], padding=1), FocusableText('<close>', handler=partial(handler, floating))], padding=3)

    def _bottom_toolbar(self):
        toolbar = super()._bottom_toolbar()
        toolbar.children.insert(0, to_container(Button('back', handler=self.back_handler)))
        return toolbar


from prompt_toolkit.layout.containers import Window, HSplit, VSplit, to_container, HorizontalAlign, WindowAlign
from prompt_toolkit.widgets import Label
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.application.current import get_app
from prompt_toolkit.key_binding import KeyBindings
from typing import Optional, Callable
from datetime import datetime


class BaseApp:

    def _bottom_toolbar(self):
        return VSplit([Label(text='<ctrl>q: quit', style='fg:ansiblue', dont_extend_width=True), Label(text='<tab/arrows>: select', dont_extend_width=True, style='fg:ansiblue'), Label(text='<enter>: confirm', dont_extend_width=True, style='fg:ansiblue')], style="bg:ansiwhite", padding=3)


class FocusableText:

    def __init__(self, text: str, align: Optional[WindowAlign] = None, ext_width=True, ext_height=True, handler: Optional[Callable] = None):
        self.handler = handler
        self.control = FormattedTextControl(text=text, focusable=True, key_bindings=self.__get_key_bindings(), show_cursor=False)
        def style():
            if get_app().layout.has_focus(self):
                return 'bg:ansired'
            return 'class:label'
        self.window = Window(
            self.control,
            style=style,
            align=align,
            height=1,
            dont_extend_width=not ext_width,
            dont_extend_height=not ext_height,
        )

    def set_handler(self, handler: Callable):
        self.handler = handler
        self.control.key_bindings = self.__get_key_bindings()

    def __get_key_bindings(self):
        kb = KeyBindings()

        @kb.add(" ")
        @kb.add("enter")
        def _(event):
            if self.handler is not None:
                self.handler()

        return kb

    def __pt_container__(self):
        return self.window


def date_formatter(timestamp: float) -> str:
    return datetime.fromtimestamp(timestamp).strftime('%c')


from __future__ import annotations
from dataclasses import dataclass, field, asdict
from typing import Any, NewType, List, Dict, Optional
from abc import ABC
from tmt.info import __version__
import typing
import json



Timestamp = NewType('Timestamp', int)

@dataclass
class BaseJsonDataclass(ABC):
    @classmethod
    def from_dict(cls, d: Dict[str, Any]):
        init_dict = {}
        for key, val in typing.get_type_hints(cls).items():
            field = cls.__solve_type(val, key, d)
            init_dict[key] = field
        return cls(**init_dict)

    @classmethod
    def from_json(cls, s: str):
        return cls.from_dict(json.loads(s))

    @staticmethod
    def __solve_type(t, key, d):
        args = typing.get_args(t)
        if len(args) == 0:
            if t is Timestamp:
                return d.get(key)
            elif issubclass(t, BaseJsonDataclass):
                return BaseJsonDataclass.init_subclass(t, key, d)
            else:
                return d.get(key)
        if BaseJsonDataclass.__is_optional_type(args):
            if d.get(key):
                return BaseJsonDataclass.__solve_type(args[0], key, d)
            return None

        container = typing.get_origin(t)
        if len(args) == 1:  # in our case, this is a list
            arg = args[0]
            arg_args = typing.get_args(arg)

            if len(arg_args) == 0:
                return container(arg.from_dict(e) if hasattr(arg, 'from_dict') else arg(e) for e in d.get(key))
            else:
                return container(BaseJsonDataclass.__solve_type(arg, 'l', {'l': e}) for e in d.get(key))
        else:
            raise ValueError('There might be a Tuple or some kind of container which is not a list somewhere in the models. This is not supported yet')

    @staticmethod
    def init_subclass(subcls, key: str, kvs: Dict):
        if key in kvs:
            return subcls.from_dict(kvs[key])
        return None

    @staticmethod
    def __is_optional_type(args):
        return len(args) > 1 and typing.get_origin(args[1]) is None

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)


@dataclass
class Result(BaseJsonDataclass):
    entry_id: str
    name: str
    path: str


@dataclass
class Metric(BaseJsonDataclass):
    entry_id: str
    name: str
    value: float


@dataclass
class Entry(BaseJsonDataclass):
    id: str
    name: str
    args: str
    date_created: Timestamp
    local_results_path: str
    local_snapshot_path: str = ""
    date_saved: Optional[Timestamp]  = None
    metrics: List[Metric] = field(default_factory=list)
    other_runs: List[Entry] = field(default_factory=list)
    results: List[Result] = field(default_factory=list)
    version: str = __version__